'============================================================
' IO12 tunnel
'
' This BCL application supports upto seven Barionet IO12 extensions.
' For Modbus address change you have to start the page add_main.html 
' (if this cob file is loaded !) in your browser i.e. http://192.168.1.10/add_main.html
' (there is also a help page add_help.html)
' Note, all fields in the upper section must be filled!
'
' The application will tunnel the IO12 input state to the according remote IO12 output (on remote Barionet).
'
' The serial number you can find on the label on the backside of your 
' Barionet extension unit (on IO12 below the barcode).
' Serial No: i.e.  1  119   ;   R272,R273 and R275 usually not used = 0   ;
' R274 = should be the modbus address ( i.e. 253) 
'
' The Io register numbers for each extension unit depend from the type of the connected extensions and the Modbus addressing
' and are displayed on the integrated Modbus addressing page.
' Currently are the addresses 255, 254, 253, 252, 251, 250, and 249 supported in this example code.
'=============================================================
'Release Notes
'
'version 1.00
'16.01.14 - SG - initial version for 7 IO12 extension units
'version 1.01
'17.01.14 - SG - added error relay when network communication existing, then ON
'version 1.02
'03.02.14 - SG - fixed counter reset in the Modbus addressing tool device table
'=============================================================
DIM Com$(25)
DIM a,b,c,d,i,n,o,u,x,k,j,uu,cp	 		 	
DIM cgiadr,adr,mbreg,mbreg_h,mblen,mbok
DIM mlen,coil, opened, short
DIM j2000,j3000,i1000,init,err,crc,tmp,ioval
DIM m$(50)
DIM mr$(50)
DIM mx$(128)
DIM mw(6)
DIM _CGI_$,c$
DIM sn1, sn1l
DIM sn2, sn2l
DIM r272, r272l
DIM r273, r273l
DIM r274, r274l
DIM r275, r275l
DIM sl$,name$(20),msg(50),cmd$(20),p$,s$
DIM rport, rhost$(20),port1,err_out,show_io
DIM start_in(7), in_all
DIM dev(7), fw(7), s_n1(7), s_n2(7),set(20),iov_old(7)



' ------------------------------------------------------------------------------
' INFO: setup memory usage
' ------------------------------------------------------------------------------
' offset   parameter
' ------------------------------------------------------------------------------
' 512..513 Tunnel UDP port (W)
' 514..517 Remote IP address (4B)
' 518	   Show local IO12 output state (Y/N) (1B)
' 519      Error Action (B)

' ------------------------------------------------------------------------------
' INFO: File handle usage
' ------------------------------------------------------------------------------
' handle	type	application
' ------------------------------------------------------------------------------
'   0     RS485 #2 Serial Tunnel
'   1     STP   #Read SETUP
'   1     UDP   #1 UDP listen/send 
' ------------------------------------------------------------------------------

name$ = sprintf$("%1A",1)
SYSLOG "Start Barionet IO12 IO tunnel on "+name$ + " v1.02", 1




' ------/------------------------------------------------------------------------
' INIT: Get setup variables
' -------------------------------------------------------------------------------

 COM$= "STP:512"
 OPEN COM$ as 1
 READ 1, set$
 Close 1
 port1= MIDGET(set$,1,2)	' 512 IO12 Tunnel UDP port
 a= MIDGET(set$,3,1)		' 514 IO12 Tunnel remote IP address byte 1
 b= MIDGET(set$,4,1)		' 515 IO12 Tunnel remote IP address byte 2
 c= MIDGET(set$,5,1)		' 516 IO12 Tunnel remote IP address byte 3
 d= MIDGET(set$,6,1)		' 517 IO12 Tunnel remote IP address byte 4
 ipa$ = STR$(a)+"."+STR$(b)+"."+STR$(c)+"."+STR$(d)
 show_io= MIDGET(set$,7,1)	' 518 Show local IO12 output state marker
 err_out= MIDGET(set$,8,1)	' 519 Error indicator
 SYSLOG "Setup READ OK",2
 SYSLOG "IO12 Tunnel to   "+ipa$+":"+STR$(port1), 2
 SYSLOG " Error Action on Barionet register "+STR$(err_out), 2


' ------------------------------------------------------------------------------
' INIT: Application enabling
' ------------------------------------------------------------------------------
COM$ = "COM:19200,E,8,1,485:2"	' RS-485 interface for relay extension
OPEN COM$ as 0

_CGI_$ = ""
ON CGI GOSUB 5000
SYSLOG "CGI handler enabled", 1  

'------------------------------------------------------------------------------------

'------------------------------------------------------------------
2 'read initial status of Extension devices
' -----------------------------------------------------------------

init=1

in_all=111
FOR c = 0 TO 6
     dev(c)=0						' resets values to 0
     fw(c)=0						' resets values to 0
     s_n1(c)=0						' resets values to 0
     s_n2(c)=0						' resets values to 0
     iov_old(c)=0
     IOCTL 761+(c*2),0					' writes to Barionet register 761 - 780
     IOCTL 762+(c*2),0
     IOCTL 301+c,0					' Error indicator for Modbus communication to IO12 module
     start_in(c)=0
NEXT c

mbreg_h=1						' high register byte
FOR c = 0 TO 6
mbreg_h=1						
mbreg=0							' reads firmware version + serial number
mblen=6							' number of bytes to read
adr=255-c
 GOSUB 200
 IF mbok=1 THEN
   IOCTL 301+c,1	
   dev(c)=(256*MIDGET(mr$,4,1)) + MIDGET(mr$,4+1,1)	' reads device type
   fw(c)=(256*MIDGET(mr$,6,1)) + MIDGET(mr$,6+1,1)	' reads firmware version
   s_n1(c)=MIDGET(mr$,9,1)				' reads first part of Serial number
   s_n2(c)=MIDGET(mr$,11,1)				' reads second part of serial number   
   
  IF dev(c)=179 THEN 					' 179 = 0xB3 for IO12 device
     init=0
     start_in(c)=in_all+(12*c) : IOCTL 761+(c*2),start_in(c)
     IOCTL 762+(c*2),start_in(c)+11
  ENDIF


 ENDIF
NEXT c

IF init=1 THEN GOTO 2

'--------------------------------------------------------------------------------------
IF port1>0 THEN						' Check IF application 1's IP(A) and port are not zero
  OPEN "UDP:0.0.0.0:"+STR$(port1) as 1
  ELSE
  SYSLOG "Tunnel disabled", 3
  Goto 9999
ENDIF
SYSLOG "UDP port = "+STR$(port1)
j3000=_TMR_(0)

mbreg_h=0
mblen=1
mbreg=1							' Modbus register for digital inputs

'-------------------- start main loop , read and copy analog input value ------------------------------------------------
10
IF _TMR_(0)-j3000>3000 THEN temp=1
FOR c = 0 TO 6
  adr=255-c						' set Modbus address for selected IO12
  IF dev(c)=179 THEN					' testing for IO12
    GOSUB 200						' send modbus read request and read, will return with response and mbok=1
    IF AND(mbok=1, a<>0) THEN				' if modbus response received
	mba=MIDGET(mr$,1,1)
	ioval=(256*MIDGET(mr$,4,1))+MIDGET(mr$,4+1,1)
	IOCTL 301+c,1					' Error indicator for Modbus communication to IO12 module
        SYSLOG "value="+SPRINTF$("%-0x",ioval)
	IF OR(ioval<>iov_old(c),temp=1) THEN
	  Write 1, mr$,7,ipa$,port1
        ENDIF
	iov_old(c)=ioval
    ENDIF
    DELAY 0
  ENDIF
GOSUB 30
     
NEXT c      

IF temp=1 THEN j3000=_TMR_(0) : temp=0

DELAY 0
IF init=1 THEN
   GOTO 2						' initialize again when Modbus parameters are changed
     ELSE 
     GOTO 10						' otherwise perform normal main loop
ENDIF							' end main loop


9999 END						

'----------------------------------------------------
30    ' read incoming network message and set outputs
'-----------------------------------------------------
IF err_out>0 THEN
   IF _TMR_(0)-j4000>4000 THEN 				' switching error relay when nothing received for 4 sec
	IOCTL err_out,0
	SYSLOG "Barionet Communication Error !!"
	ELSE
	  IOCTL err_out,1
   ENDIF
ENDIF   
IF LASTLEN(1) = -7 THEN
      rport=RMTPORT(1)
      rhost$=RMTHOST$(1)
      READ 1,cmd$
      b=LEN(cmd$)
      SYSLOG " Incoming Command from: "+rhost$+":" + STR$(rport)+ "( "+STR$(b)+" byte)" 
31    d=MIDGET(cmd$,2,1)
	 IF d = 3 THEN					' check second received byte = 0x03
	    IF err_out>0 THEN j4000=_TMR_(0)
      	    mba=MIDGET(cmd$,1,1)
	    iov1=MIDGET(cmd$,4,1)
	    iov2=MIDGET(cmd$,5,1)
            GOSUB 800
	    IF show_io=1 THEN SYSLOG "Show_IO = 1 " : GOSUB 999
 	    b=LEN(cmd$)					' coyies the second command in a temporary buffer
	    cmd$=MID$(cmd$,8)				' coyies the second command in a temporary buffer
	    b=LEN(cmd$)
      	    IF b>=3 THEN				' if second command found ...
		GOTO 31					' executes the further commands
		ELSE
		  cmd$= ""
            ENDIF
         ENDIF
    ENDIF
RETURN


	
'-------------------------------------------------------------------------------
200 'modbus read for intial relay state and for input state
'-------------------------------------------------------------------------------
  MIDSET m$,1,1,adr			' set address
  MIDSET m$,2,1,3			' read registers command
  MIDSET m$,3,1,mbreg_h			' hi address
  MIDSET m$,4,1,mbreg			' low address
  MIDSET m$,5,1,0			' hi number
  MIDSET m$,6,1,mblen			' lo number
  'MIDSET m$,7,1,0			' crc value
  'MIDSET m$,8,1,0			' crc value
  mlen=8
  GOSUB 1000				' make modbus crc
  GOSUB 700 				' read out stuff from interface, if line noise (unlikely)
  WRITE 0,m$,mlen			' send out the block
  GOSUB 2000
  GOSUB 300				' wait for answer
RETURN


'========================================================================
' Modus read - read back answer from modbus in mr$, return status with it
'========================================================================
300
j2000=_TMR_(0)

400
READ 0,mr$,10		 		' try to read
IF AND(_TMR_(0)-j2000<80,LASTLEN(0)<=0) THEN GOTO 400 'wait upto 100 ms for response
mlen=LASTLEN(0)
IF mlen>3 THEN
    d=1					' set marker for debugging info
    GOSUB 2000				' send debugging SYSLOG
    crc=&HFFFF
   FOR i1000=1 TO mlen-2
   crc=xor(crc,MIDGET(mr$,i1000,1))
	IF AND(crc,1) THEN crc=xor(crc,&H14002)
	IF AND(crc,2) THEN crc=xor(crc,&H28004)
	IF AND(crc,4) THEN crc=xor(crc,&H50008)
	IF AND(crc,8) THEN crc=xor(crc,&HA0010)
	IF AND(crc,16) THEN crc=xor(crc,&H140020)
   	IF AND(crc,32) THEN crc=xor(crc,&H280040)
   	IF AND(crc,64) THEN crc=xor(crc,&H500080)
   	IF AND(crc,128) THEN crc=xor(crc,&HA00100)
   	crc=SHR(crc,8)
   	NEXT i1000		 
   	IF MIDGET(mr$,mlen-1,2)<>crc THEN
      	   SYSLOG "read: crc error"
      	   mbok=-2
           ELSE
           mbok=1
	   DELAY 10
   	ENDIF
   ELSE
   SYSLOG "no answer from modbus device " + STR$(adr), 7
   mbok=-1
   IOCTL 301+c,0	
ENDIF
RETURN


'====== read/clean noise trash from the line ===============

700 					' Read response
READ 0, mx$,1		 		' try to read
RETURN
'===========================================================

800 'Set outputs
  MIDSET m$, 1, 1, mba  'Modbus address
  MIDSET m$, 2, 1, 6    'Modbus command multiple outputs
  MIDSET m$, 3, 1, 0    'Address Hi
  MIDSET m$, 4, 1, 0	'Address Lo
  MIDSET m$, 5, 1, iov1 'Value Hi
  MIDSET m$, 6, 1, iov2 'Value Lo
  mlen = 8
  GOSUB 1000
  GOSUB 700
  WRITE 0,m$,mlen			' send out the block
  GOSUB 2001
  GOSUB 700
RETURN



'=======copy Output state to Barionet=======================
999
n=255-mba
FOR i = 0 TO 11					' for IO12 check 12 registers
    p=SHR(AND(MIDGET(cmd$,4,-2),2^i), i)
    f=111+(n*12)+i
    SYSLOG "IOCTL "+STR$(f)+","+STR$(p) 
    IOCTL f,p					' copies input registers 2xx (here 1 for close, 0 for open)
NEXT i

RETURN

'===========================================================
1000 'Modbus CRC calc
'=====================
  crc = &HFFFF
  FOR i1000 = 1 TO (mlen - 2)
    crc = XOR(crc, MIDGET(m$, i1000, 1))
    FOR j = 1 TO 8
      k = AND(crc, 1)
      crc = SHR(crc, 1)
      IF k = 1 THEN crc = XOR(crc, &HA001)
    NEXT j
  NEXT i1000        
  MIDSET m$, (mlen - 1), 2, crc
RETURN

'------------------------------------------------------------------------
2000 ' SYSLOG DEBUGGING for MODBUS communication
'------------------------------------------------------------------------
  IF d=1 THEN						' d=1 is "read" , d=0 is "write"
    p$="read "+STR$(adr)+SPRINTF$(": %d bytes: ",mlen)	' prepare IO12/X8 response for SYSLOG
    FOR o=1 TO mlen 
	p$=p$+SPRINTF$(" %02x",MIDGET(mr$,o,1)) 	' prepare IO12/X8 response/command for SYSLOG
    NEXT o
    ELSE
       p$="write "+STR$(adr)+SPRINTF$(": %d bytes: ",mlen)	' prepare IO12/X8/R6 response for SYSLOG
       FOR o=1 TO mlen 
	 p$=p$+SPRINTF$(" %02x",MIDGET(m$,o,1)) 	' prepare IO12/X8/R6 response/command for SYSLOG
       NEXT o
  ENDIF
  SYSLOG p$, 6						' syslog IO12/X8 Modbus response string
  d=0
RETURN

2001 ' SYSLOG DEBUGGING for MODBUS SET (06) command
'------------------------------------------------------------------------
  
p$="SET OUTPUT write "+STR$(mba)+SPRINTF$(": %d bytes: ",mlen)	' prepare IO12 response for SYSLOG
FOR o=1 TO mlen 
   p$=p$+SPRINTF$(" %02x",MIDGET(m$,o,1)) 	' prepare IO12 response/command for SYSLOG
NEXT o
SYSLOG p$, 6					' syslog IO12 Modbus response string

RETURN



'------------------------------------------------------------------------
5000 'CGI Handler to change Modbus address settings
'------------------------------------------------------------------------
  c$ = _CGI_$
  _CGI_$ = ""
  WRITE -1,"HTTP/1.1 200 OK\r\n\r\n",0
   _CGI_$ = "*"
  IF LEN(c$) > 0 THEN
    SYSLOG "CGI : " + c$, 1
    sn1 = INSTR(1, c$, "sn1=")
    sn2 = INSTR(1, c$, "sn2=")
    r272 = INSTR(1, c$, "r272=")
    r273 = INSTR(1, c$, "r273=")
    r274 = INSTR(1, c$, "r274=")
    r275 = INSTR(1, c$, "r275=")
    IF AND(sn1 > 0, sn2 > 0, r272 > 0, r273 > 0, r274 > 0, r275 > 0) THEN
      sn1 = sn1 + 4
      sn2 = sn2 + 4
      r272 = r272 + 5
      r273 = r273 + 5
      r274 = r274 + 5
      r275 = r275 + 5
      tmp = INSTR(sn1, c$, "&")
      IF tmp > 0 THEN
        sn1l = tmp - sn1
      ELSE
        sn1l = LEN(c$) - sn1
      ENDIF
      tmp = INSTR(sn2, c$, "&")
      IF tmp > 0 THEN
        sn2l = tmp - sn2
      ELSE
        sn2l = LEN(c$) - sn2
      ENDIF
      tmp = INSTR(r272, c$, "&")
      IF tmp > 0 THEN
        r272l = tmp - r272
      ELSE
        r272l = LEN(c$) - r272
      ENDIF
      tmp = INSTR(r273, c$, "&")
      IF tmp > 0 THEN
        r273l = tmp - r273
      ELSE
        r273l = LEN(c$) - r273
      ENDIF
      tmp = INSTR(r274, c$, "&")
      IF tmp > 0 THEN
        r274l = tmp - r274
      ELSE
        r274l = LEN(c$) - r274 + 1
      ENDIF
      tmp = INSTR(r275, c$, "&")
      IF tmp > 0 THEN
        r275l = tmp - r275
      ELSE
        r275l = LEN(c$) - r275 + 1
      ENDIF
      IF AND(sn1l > 0, sn2l > 0, r272l > 0, r273l > 0, r274l > 0, r275l > 0) THEN
        sn1  = VAL(MID$(c$, sn1, sn1l))
        sn2  = VAL(MID$(c$, sn2, sn2l))
        r272 = VAL(MID$(c$, r272, r272l))
        r273 = VAL(MID$(c$, r273, r273l))
        r274 = VAL(MID$(c$, r274, r274l))
        r275 = VAL(MID$(c$, r275, r275l))
        MIDSET m$, 1, 1, 0     				' Broadcast
        MIDSET m$, 2, 1, 69    				' Set config
        MIDSET m$, 3, 1, sn1   				' Serial number, part 1
        MIDSET m$, 4, 1, sn2   				' Serial number, part 2        
        MIDSET m$, 5, 1, sn1   				' Serial number, part 1 again
        MIDSET m$, 6, 1, sn2   				' Serial number, part 2 again
        MIDSET m$, 7, 1, r272  				' Relay power on defaults
        MIDSET m$, 8, 1, r273  				' Serial interface mode
        MIDSET m$, 9, 1, r274  				' Modbus address
        MIDSET m$, 10, 1, r275 				' Reserved  
        mlen = 12
	GOSUB 1000
	GOSUB 700  
	WRITE 0, m$, mlen
	GOSUB 2000		 			' will SYSLOG Modbus command for debugging
	GOSUB 700
	SYSLOG "Set config for device " + STR$(sn1) + "-" + STR$(sn2), 1
	ELSE
	  SYSLOG "Invalid CGI request (Missing Parameter Values)", 1
	ENDIF
      ELSE
	tmp = INSTR(1, c$, "cgiadr=")			' command will be used for future features
	IF AND(tmp > 0, LEN(c$) > 4) THEN
	  tmp = tmp + 4
          tmp = VAL(MID$(c$, tmp, LEN(c$) - tmp + 1))
          cgiadr = tmp
          ELSE
            SYSLOG "Invalid CGI request (Missing Parameters Name)", 1
	ENDIF    
      ENDIF
  ELSE
    SYSLOG "Invalid CGI request (No Command)", 1
ENDIF
init=1
RETURN
